from zeep.transports import Transport
from zeep import Client, Settings
from requests import Session
from requests.auth import HTTPBasicAuth
import sys
import time
from datetime import datetime, timedelta


class CDROnDemand():
    class ProgressBar():
        progress_bar_scale = .7
        progress_bar_length = int(100 * progress_bar_scale)
        completed_bar_char = "█"
        incomplete_bar_char = "▒"
        completion_message = "Completed!"

        def __init__(self, total):
            print()
            self.total = total

        def progress(self, current):
            # print out our progress bar
            percent_complete = round(current / self.total * 100)
            bars_of_completion = int(percent_complete * self.progress_bar_scale)
            bars_to_be_completed = self.progress_bar_length - bars_of_completion
            print(
                f'\r{self.completed_bar_char * bars_of_completion}{self.incomplete_bar_char * bars_to_be_completed} {percent_complete}%',
                end='')

    def __init__(self,ipaddr,username,passwd):
        # set DIME WSDL file from URL
        dime_wsdl = f'https://{ipaddr}:8443/logcollectionservice/services/DimeGetFileService?wsdl'

        # create a SOAP client session
        session = Session()

        # avoid certificate verification by default and setup session
        session.verify = False
        session.auth = HTTPBasicAuth(username, passwd)
        transport = Transport(session=session, timeout=10)
        settings = Settings(strict=False, xml_huge_tree=True)

        # Create client and soap service
        client_dime = Client(dime_wsdl, settings=settings,
                                transport=transport)  # , plugins = plugin )
        self.dime = client_dime.create_service(
            '{http://cisco.com/ccm/serviceability/soap/LogCollection/GetFile/}GetFileBinding',
            dime_wsdl)

        # set CDRonDemand WSDL file from URL for File Transfer
        cdr_on_demand_wsdl = f'https://{ipaddr}:8443/CDRonDemandService2/services/CDRonDemandService?wsdl'

        # Create client and soap service
        client_cdr_on_demand = Client(cdr_on_demand_wsdl, settings=settings,
                                      transport=transport)  # , plugins = plugin )
        self.cdr_on_demand = client_cdr_on_demand.create_service(
            '{http://schemas.cisco.com/ast/soap}CDRonDemandSoapBinding',
            cdr_on_demand_wsdl)

    # File request from CM via DIME
    def get_file(self, filepath):
        # check if file path includes /var/log as required if not assume /var/log/active
        if filepath.find('/var/log') == -1:
            filepath = f"/var/log/active/{filepath}"

        try:
            file_contents = self.dime.GetOneFile(filepath).attachments[0].content.decode('UTF-8')

        except Exception as err:
            print(f"Error pulling {filepath} from CUCM - {err}", file=sys.stderr)
            return False

        else:
            return file_contents

    # get CDR files to get (list or single file with return_type = dict|string|dataframe
    def get_cdr_file(self, filename, return_type="list"):
        cdr_return = {'cdr': { 'columns': [], 'types': [], 'contents': [] },
                      'cmr': { 'columns': [], 'types': [], 'contents': [] } }
        cdr_files = {}

        # if filename is string (single file) then wrap into a list to process
        if isinstance(filename, str):
            filename = [filename]
        pb = self.ProgressBar(len(filename))

        # get each file in list and process into dict formatting
        x = 1
        start = time.time()
        for file in filename:
            pb.progress(x)
            x += 1
            try:
                file_contents = self.get_file(f"/var/log/active/cm/cdr_repository/processed/{file.split('_')[3][:8]}/{file}").replace('"', '').splitlines()

            except Exception as err:
                print(f"Error retrieving filename {file} - {err}", file=sys.stderr)

            else:
                # parse out file lines
                if file.find("cdr_") > -1:
                    if len(cdr_return['cdr']['columns']) == 0:
                        cdr_return['cdr']['columns'] = file_contents[0].split(',')
                        cdr_return['cdr']['types'] = file_contents[1].split(',')

                    for line in file_contents[2:]:
                        y = 0
                        row = []
                        line = line.split(',')
                        while y < len(line):
                            if not cdr_return['cdr']['types'][y] == '':
                                if cdr_return['cdr']['types'][y] == 'INTEGER':
                                    try:
                                        row.append(int(line[y]))
                                    except:
                                        row.append(0)
                                else:
                                    row.append(line[y])
                            else:
                                row.append('')
                            y += 1

                        cdr_return['cdr']['contents'].extend([row])

                elif file.find("cmr_") > -1:
                    if len(cdr_return['cmr']['columns']) == 0:
                        cdr_return['cmr']['columns'] = file_contents[0].split(',')
                        cdr_return['cmr']['types'] = file_contents[1].split(',')
                    for line in file_contents[2:]:
                        y = 0
                        row = []
                        line = line.split(',')
                        while y < len(line):
                            if not cdr_return['cmr']['types'][y] == '':
                                if cdr_return['cmr']['types'][y] == 'INTEGER':
                                    try:
                                        row.append(int(line[y]))
                                    except:
                                        row.append(0)
                                else:
                                    row.append(line[y])
                            else:
                                row.append('')
                            y += 1

                        cdr_return['cmr']['contents'].extend([row])

        print(f"{len(filename)} files in {time.time() - start}sec.")
        if return_type.lower == "str" or return_type.lower() == "string":
            cdr_string = ",".join(cdr_return['cdr']['columns']) + '\n'
            for i in cdr_return['cdr']['contents']:
                cdr_string = cdr_string + ",".join(i) + '\n'

            cmr_string = ",".join(cdr_return['cmr']['columns']) + '\n'
            for i in cdr_return['cmr']['contents']:
                cmr_string = cmr_string + ",".join(i) + '\n'

            return cdr_string

        elif return_type.lower() == "list":
            return { 'cdr': cdr_return['cdr']['contents'],
                     'cmr': cdr_return['cmr']['contents'] }

        else:
            return cdr_return

    # List CDR files used between start/stop epoch time
    def list_cdr_files(self, cdr_start, cdr_stop):
        cdr_start = datetime.strptime(cdr_start,'%Y%m%d%H%M')
        cdr_stop = datetime.strptime(cdr_stop,'%Y%m%d%H%M')
        file_list = []
        while cdr_start < cdr_stop:
            try:
                files = self.cdr_on_demand.get_file_list(cdr_start.strftime('%Y%m%d%H%M'), (cdr_start + timedelta(hours=1)).strftime('%Y%m%d%H%M'), True)
            except Exception as err:
                print(f"Error getting file names - {err}", file=sys.stderr)
            else:
                for file in files:
                    file_list.append(file)
            cdr_start = cdr_start + timedelta(hours=1)

        return file_list


# cm = CMFileCollector.Connect("10.45.50.110","API","API")
# #File path = /var/log/<1>/<2> where CLI = file view <1> <2>
# cdr = cm.dime.GetOneFile('/var/log/active/cm/cdr_repository/processed/20210112/cmr_Nucor-EnterpriseCluster-01_06_202101122006_158160')
# print(cdr.attachments[0].content)
#
# b'"cdrRecordType","globalCallID_callManagerId","globalCallID_callId","nodeId","directoryNum","callIdentifier","dateTimeStamp","numberPacketsSent","numberOctetsSent","numberPacketsReceived","numberOctetsReceived","numberPacketsLost","jitter","latency","pkid","directoryNumPartition","globalCallId_ClusterID","deviceName","varVQMetrics","duration","videoContentType","videoDuration","numberVideoPacketsSent","numberVideoOctetsSent","numberVideoPacketsReceived","numberVideoOctetsReceived","numberVideoPacketsLost","videoAverageJitter","videoRoundTripTime","videoOneWayDelay","videoReceptionMetrics","videoTransmissionMetrics","videoContentType_channel2","videoDuration_channel2","numberVideoPacketsSent_channel2","numberVideoOctetsSent_channel2","numberVideoPacketsReceived_channel2","numberVideoOctetsReceived_channel2","numberVideoPacketsLost_channel2","videoAverageJitter_channel2","videoRoundTripTime_channel2","videoOneWayDelay_channel2","videoReceptionMetrics_channel2","videoTransmissionMetrics_channel2","localSessionID","remoteSessionID","headsetSN","headsetMetrics"\nINTEGER,INTEGER,INTEGER,INTEGER,VARCHAR(50),INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,UNIQUEIDENTIFIER,VARCHAR(50),VARCHAR(50),VARCHAR(129),VARCHAR(600),INTEGER,VARCHAR(10),INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,VARCHAR(600),VARCHAR(600),VARCHAR(10),INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,VARCHAR(600),VARCHAR(600),VARCHAR(128),VARCHAR(128),VARCHAR(129),VARCHAR(1024)\n2,6,11269506,6,"+193668

# files = cm.list_cdr_files('202101130700','202101130800') #start/stop datetime year,month,day,hour,minute
# file = cm.get_cdr_file(files[0])



import requests
from requests import Session
from requests.auth import HTTPBasicAuth
import sys
import xmltodict
import re
import os
import json
import urllib
from .AST import *
from .Serviceability import *
from .DIME import *
from .Logs import Logs
from .CDR import CDR
from . import AXL
from . import Risport
from . import UDS
import socket

requests.packages.urllib3.disable_warnings()
try:
    requests.packages.urllib3.contrib.pyopenssl.util.ssl_.DEFAULT_CIPHERS += ':HIGH:!DH:!aNULL'
    requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += ':HIGH:!DH:!aNULL'
except AttributeError:
    # no pyopenssl support used / needed / available
    pass


class Connect:
    # initialize object set system type and build DB connectors as needed
    def __init__(self, ipaddr=None, username=None, passwd=None, version=None, wsdl=None):
        self.system_type = "cucm"
        self.ipaddr = ipaddr
        self.axl = False
        self.risport = False
        self.auth = HTTPBasicAuth(username, passwd)

        try:
            self.uds = UDS.Connect(ipaddr, username, passwd)
            self.version = self.uds.get("version")['@version']
        except Exception as err:
            print(f"Could not determine CUCM version via UDS: {err}")
            self.version = False

        if self.open_port(ipaddr, 8443) is False:
            raise Exception (f"Connection Error: {ipaddr}:8443 not reachable or open. Is this CUCM?")
        if self.query("SELECT COUNT(*) FROM processnode") is False:
            raise Exception (f"Connection Error: AXL request not valid. Improper credentials?")

        risport = Risport.Connect(ipaddr, username, passwd)
        self.risport = risport.client

        axl = AXL.Connect(ipaddr, username, passwd, '.'.join(self.version.split('.')[0:2]))
        self.axl_client = axl.client
        self.axl = axl

        self.ast = AST.Connect(ipaddr, username, passwd)
        self.cluster = self.ast.cluster
        if self.cluster:
            self.serviceability = Serviceability.Connect(ipaddr, username, passwd, self.cluster['nodes'])
        else:
            self.serviceability = Serviceability.Connect(ipaddr, username, passwd)

        try:
            self.dime = DIME.Connect(ipaddr, username, passwd)
        except Exception as err:
            print(f"Could not connect to DIME: {err}")

        # self.logs = Logs(ipaddr, username, passwd)
        self.cdr = CDR(ipaddr, username, passwd)

        print("Connected.")

    @staticmethod
    def open_port(ip, port, return_object=[]):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(3)
        try:
            s.connect((ip, int(port)))
            s.close()
            return_object.append(str(port))
            return port
        except Exception as e:
            s.close()
            return False

    # Function to query SQL via AXL API to CUCM
    def query(self, sql_statement):
        if self.version:
            version = self.version
        else:
            version = "8.5"
        result_dict = []
        row_info = ["", None, None]
        newsql_statement = sql_statement
        while row_info[1] is None or int(row_info[1]) > len(result_dict):
            if re.search("select", sql_statement.lower()):
                execute_sql = "executeSQLQuery"
            else:
                execute_sql = "executeSQLUpdate"

            if row_info[1] is not None and int(row_info[1]) > 0:
                newsql_statement = re.sub("[sS][eE][lL][eE][cC][tT] ",
                                          f"SELECT SKIP {len(result_dict)} LIMIT {row_info[2]} ", sql_statement)

            payload = f"""<?xml version=\"1.0\" encoding=\"UTF-8\"?>
                            <soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\" xmlns:ns=\"http://www.cisco.com/AXL/API/{'.'.join(version.split(".")[0:2])}\">
                              <soapenv:Header/>
                              <soapenv:Body>
                                <ns:{execute_sql}>
                                  <sql>
                                    {newsql_statement}
                                  </sql>
                                </ns:{execute_sql}>  
                              </soapenv:Body>
                            </soapenv:Envelope>"""
            try:
                # send API request
                response = requests.request("POST", f'https://{self.ipaddr}:8443/axl/',
                                            headers={'Content-Type': 'text/xml'},
                                            auth=self.auth,
                                            data=payload,
                                            verify=False)
            except Exception as err:
                print(f"Error with AXL DB request: {err}", file=sys.stderr)
                return False
            else:
                # If request was good then return response values
                if response.status_code == 200:
                    # normalize output to always show array of rows
                    if re.search('ns:executeSQLUpdateResponse', response.text):
                        result_dict.extend([xmltodict.parse(response.text)['soapenv:Envelope']['soapenv:Body'][
                                                'ns:executeSQLUpdateResponse']['return']])
                        break

                    elif xmltodict.parse(response.text)['soapenv:Envelope']['soapenv:Body']['ns:executeSQLQueryResponse'][
                        'return'] is not None:
                        if type(xmltodict.parse(response.text, dict_constructor=dict)['soapenv:Envelope']['soapenv:Body'][
                                    'ns:executeSQLQueryResponse']['return']['row']) is not list:
                            result_dict.extend([xmltodict.parse(response.text, dict_constructor=dict)['soapenv:Envelope'][
                                                    'soapenv:Body']['ns:executeSQLQueryResponse']['return']['row']])
                        else:
                            result_dict.extend(
                                xmltodict.parse(response.text, dict_constructor=dict)['soapenv:Envelope']['soapenv:Body'][
                                    'ns:executeSQLQueryResponse']['return']['row'])

                        if row_info[1] is None or int(row_info[1]) <= len(result_dict):
                            break
                    else:
                        return []

                elif re.search("Query request too large. Total rows matched: ", response.text):
                    row_info = re.sub("[^\d:]", '',
                                      xmltodict.parse(response.text, dict_constructor=dict)['soapenv:Envelope'][
                                          'soapenv:Body']['soapenv:Fault']['faultstring']).split(":")
                    print(f"Multiple queries needed for {row_info[1]} rows.")

                else:
                    print(f'Requests Error: {response}', file=sys.stderr)
                    print(f'\t{sql_statement}', file=sys.stderr)
                    print(f'\t{response.text}', file=sys.stderr)
                    return False

        return result_dict

    # risport query for device registration
    def registration_status(self, device_list=None, return_format="dict"):
        if self.risport is False:
            print("risport client not available.", file=sys.stderr)
            return False
        # If no list given then get all devices to list
        if device_list is None:
            device_list = self.query(f"SELECT name AS device FROM device WHERE tkclass = 1")
        elif type(device_list) == str:
            device_list = [device_list]

        i = 0
        risport_response = {}
        while i < len(device_list):
            # Build and execute the risport request object
            state_info = ''

            # Build Search Criteria
            risport_search_criteria = {
                'MaxReturnedDevices': '1000',
                'DeviceClass': 'Any',
                'Model': '255',
                'Status': 'Any',
                'NodeName': '',
                'SelectBy': 'Name',
                'Protocol': 'Any',
                'DownloadStatus': 'Any',
                'SelectItems': {
                    'item': []
                }
            }
            # One or more specific devices can be retrieved by replacing * with
            # the device name in multiple items
            device_count = 0
            while device_count < 1000 and (device_count + i) < len(device_list):
                if isinstance(device_list[device_count + i], dict):
                    risport_response[device_list[device_count + i]['device']] = {'description': '', 'protocol': '',
                                                                                 'type': '', 'firmware': '', 'Node': '',
                                                                                 'IPAddress': '', 'dirn': '',
                                                                                 'status': 'Never'}
                    risport_search_criteria['SelectItems']['item'].append(
                        {'Item': device_list[device_count + i]['device']})

                elif isinstance(device_list[device_count + i], str):
                    risport_response[device_list[device_count + i]] = {'description': '', 'protocol': '', 'type': '',
                                                                       'firmware': '', 'Node': '', 'IPAddress': '',
                                                                       'dirn': '', 'status': 'Never'}
                    risport_search_criteria['SelectItems']['item'].append({'Item': device_list[device_count + i]})

                device_count = device_count + 1

            i += device_count

            # Get reg statuses
            try:
                resp = self.risport.selectCmDevice(state_info, risport_search_criteria)
            except Fault as err:
                print(f'Zeep error: selectCmDevice: {err}', file=sys.stderr)
                return False

            # format results for easy parsing
            for node in resp['SelectCmDeviceResult']['CmNodes']['item']:
                if node['ReturnCode'] != 'Ok':
                    continue

                for device in node['CmDevices']['item']:
                    risport_response[device['Name']] = {}
                    risport_response[device['Name']]['name'] = device['Name']
                    risport_response[device['Name']]['node'] = node['Name']
                    ipaddresses = device['IPAddress']
                    risport_response[device['Name']]['description'] = device['Description'] if device[
                                                                                                   'Description'] is not None else ''
                    risport_response[device['Name']]['protocol'] = device['Protocol'] if device[
                                                                                             'Protocol'] is not None else ''
                    risport_response[device['Name']]['type'] = device['DeviceClass'] if device[
                                                                                            'DeviceClass'] is not None else ''
                    risport_response[device['Name']]['firmware'] = device['ActiveLoadID'] if device[
                                                                                                 'ActiveLoadID'] is not None else ''
                    risport_response[device['Name']]['IPAddress'] = ipaddresses['item'][0]['IP'] if ipaddresses else ''

                    if device['LinesStatus'] is None:
                        risport_response[device['Name']]['dirn'] = "unknown"
                        risport_response[device['Name']]['status'] = device['Status']

                    else:
                        for x in range(0, len(device['LinesStatus']['item'])):
                            if x == 0:
                                risport_response[device['Name']]['dirn'] = device['LinesStatus']['item'][x][
                                    'DirectoryNumber']
                                risport_response[device['Name']]['status'] = device['LinesStatus']['item'][x]['Status']
                                break

        # if default dict format then return dict
        if return_format == "dict":
            return risport_response

        # else if return format is list then loop through to convert dict based on device name to list of results
        elif return_format == "list":
            risport_list = []
            for item in risport_response:
                if "name" in risport_response[item]:
                    risport_list.append(risport_response[item])

            return risport_list

    def update_pt(self, uuid, partition):
        try:
            dn_return = self.update.Line(uuid=uuid, newRoutePartitionName=partition)

        except Exception as err:
            print(f"Error during DN Update: ", err)
            return False

        else:
            return dn_return



import requests
from requests.auth import HTTPBasicAuth
import sys
import os
import re
import urllib
import socket
import json

requests.packages.urllib3.disable_warnings()
try:
    requests.packages.urllib3.contrib.pyopenssl.util.ssl_.DEFAULT_CIPHERS += ':HIGH:!DH:!aNULL'
    requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += ':HIGH:!DH:!aNULL'
except AttributeError:
    # no pyopenssl support used / needed / available
    pass

class Connect:
    # initialize object set system type and build DB connectors as needed
    def __init__(self,
                 ipaddr=None,
                 username=None,
                 passwd=None,
                 db=False,
                 unity_db="UnityDirDb"
                 ):

        self.system_type = "Unity Connection"
        self.ipaddr = ipaddr

        # if type= cucm then set username/password for AXL connection
        if ipaddr is None or passwd is None or username is None:
            print(f'Usage: CollabConnector.CUCX("ipaddr", "admin", "password")',
                  file=sys.stderr)
        else:
            self.username = username
            self.auth = HTTPBasicAuth(username, passwd)

            if self.open_port(ipaddr, 8443) is False:
                raise Exception(f"Connection Error: {ipaddr}:8443 not reachable or open. Is this Unity Connection?")
            if self.get("cluster") is False:
                print(f"Connection Error: AXL request not valid. Improper credentials?", file=sys.stderr)

            try:
                self.version = self.get("version")[0]['version']
            except Exception as err:
                raise f"Could not connect via vmrest: {err}"

            if db:
                try:
                    import IfxPy
                except:
                    print("""IfxPy module failed to load. If you intend to connect to an Informix DB check the linux enviroment
                            << pip3 install IfxPy >>
                            export INFORMIXDIR=/opt/IBM/Informix_Client-SDK/  
                            export CSDK_HOME=$INFORMIXDIR
                            export LIBPATH=${INFORMIXDIR}/lib:${INFORMIXDIR}/lib/cli:${INFORMIXDIR}/lib/esql:${INFORMIXDIR}/lib:${INFORMIXDIR}/bin:${INFORMIXDIR}/etc:${LIBPATH}
                            export LD_LIBRARY_PATH=$INFORMIXDIR/lib:$INFORMIXDIR/lib/cli:$INFORMIXDIR/lib/esql
                            """, file=sys.stderr)
                    self.INFORMIX = False
                else:
                    connection_string = f'CLIENT_LOCALE=en_US.57372;DB_LOCALE=en_US.57372;uid={username};pwd={passwd};SERVER=ciscounity;DATABASE={unity_db};HOST={ipaddr};SERVICE=20532;PROTOCOL=onsoctcp'
                    try:
                        self.informixConnector = IfxPy.connect(connection_string, "", "")
                    except Exception as err:
                        print(f'Could not Connect to Unity Informix DB: {err}', file=sys.stderr)
                        self.INFORMIX = False
                    else:
                        self.INFORMIX = True
            else:
                self.INFORMIX = False
            print("Connected.")

    @staticmethod
    def open_port(ip, port, return_object=[]):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(3)
        try:
            s.connect((ip, int(port)))
            s.close()
            return_object.append(str(port))
            return port
        except Exception as e:
            s.close()
            return False

    # Function query SQL via direct Informix connector
    def informix_query(self, sql_statement):
        if self.INFORMIX is False:
            print("Not connected to INFORMIX DB. Use db=True.", file=sys.stderr)
            return False

        result = []
        if re.search("select", sql_statement.lower()) and not re.search('select.*limit .*from', sql_statement.lower()):
            select_count = re.sub("[sS][eE][lL][eE][cC][tT] .* [fF][rR][oO][mM] ", "SELECT COUNT(*) AS qty FROM ",
                                 sql_statement)
            stmt = IfxPy.exec_immediate(self.informixConnector, select_count)
            return_rows = int(IfxPy.fetch_assoc(stmt)['qty'])

            request_count = 0
            while request_count < return_rows:
                request_rows = re.sub("[sS][eE][lL][eE][cC][tT] ", f"SELECT SKIP {request_count} LIMIT 4000 ",
                                      sql_statement)

                try:
                    stmt = IfxPy.exec_immediate(self.informixConnector, request_rows)
                except Exception as err:
                    print(f"SQL Error: {err}", file=sys.stderr)
                    return False
                else:
                    while True:
                        assoc = IfxPy.fetch_assoc(stmt)
                        if assoc is not False:
                            result.append(assoc)
                        else:
                            break

                request_count = len(result)

        else:
            try:
                stmt = IfxPy.exec_immediate(self.informixConnector, sql_statement)
            except Exception as err:
                print(f"SQL Error: {err}", file=sys.stderr)
                return False
            else:
                assoc = IfxPy

            while True:
                assoc = IfxPy.fetch_assoc(stmt)
                if assoc is not False:
                    result.append(assoc)
                else:
                    break

        return result

    # REST Wrapper for Unity
    def cupi_api(self, target_uri, method='GET', data={}, paging=True):
        # build URL based on URI in funcion call
        if target_uri.find("vmrest/") > -1:
            target_uri = "".join(target_uri.split("vmrest/")[1:])

        target_uri = f"https://{self.ipaddr}:8443/vmrest/{target_uri}"

        headers = {'Accept': "application/json",
                   'Content-Type': "application/json"}

        if method in ['PUT', 'POST', 'DELETE']:
            try:
                response = requests.request(method, target_uri, headers=headers, auth=self.auth,
                                            verify=False, data=json.dumps(data))
            except Exception as err:
                print(f"CXN API {method} Failure: {err} - {target_uri}", file=sys.stderr)
                return False
            else:
                result = [response.text]
        else:
            #  Add parameters to uri if needed
            if len(data) > 0:
                target_uri += f"?{urllib.parse.urlencode(data)}"

            # for pagination set the separator value if URI has get variables or not
            if target_uri.find("?") > -1:
                separator = "&"
            else:
                separator = "?"

            if method == 'GET' and paging is True:

                # get number of results fpr pagination
                try:
                    response = requests.get(f"{target_uri}{separator}pageNumber=0", headers=headers,
                                            auth=self.auth, verify=False)
                except Exception as err:
                    print(f"CXN API GET Failure: {err} - {target_uri}", file=sys.stderr)
                    return False

                # if @totals exist in response then expect more than 1 result and set max entries
                if "@total" in response.text:
                    try:
                        query_count = int(json.loads(response.text)['@total'])
                    except Exception as err:
                        pass
                else:
                    query_count = 1

                if query_count == 0:
                    return []

                # find page count if 1000 items per page for loop
                max_page = round(query_count / 1000) + 1
            else:
                max_page = 1
            current_page = 1

            # create result array
            result = []
            # loop through results getting 500 entries at a time
            while current_page <= max_page:
                try:
                    response = requests.get(f"{target_uri}{separator}rowsPerPage=1000&pageNumber={str(current_page)}",
                                            headers=headers, auth=self.auth, verify=False)
                except Exception as err:
                    print(f"\tCXN GET API Failure: {err} - {target_uri}", file=sys.stderr)
                    return False

                # confirm if valid response
                if 200 <= response.status_code <= 300:
                    query_result = json.loads(response.text)

                    # if multiple results in response loop through each and add to result array
                    if "@total" in query_result.keys():
                        for key in query_result:
                            if not key == "@total":
                                if query_result['@total'] == "1":
                                    result.append(query_result[key])
                                else:
                                    for item in query_result[key]:
                                        result.append(item)
                    else:
                        result.append(query_result)
                elif 400 <= response.status_code <= 600:
                    print(f"\tCXN GET API Failure: {response.text} - {target_uri}", file=sys.stderr)
                    return False

                else:
                    print(f"\t{str(response)}{response.text}", file=sys.stderr)
                    return False

                current_page += 1

        return result

    # wrapper for api Gets
    def get(self, target_endpoint, params={}, DtmfAccessId=None, Alias=None, query=None, sort=None):
        if DtmfAccessId or Alias or query:
            if DtmfAccessId:
                if DtmfAccessId.endswith('%') or DtmfAccessId.endswith('*'):
                    params['query'] = f'(DtmfAccessId startswith {DtmfAccessId.replace("%", "").replace("*", "")})'
                else:
                    params['query'] = f'(DtmfAccessId is {DtmfAccessId})'
            elif Alias:
                if Alias.endswith('%') or Alias.endswith('*'):
                    params['query'] = f'(Alias startswith {Alias.replace("%", "").replace("*", "")})'
                else:
                    params['query'] = f'(Alias is {Alias})'
            elif query:
                params['query'] = query

        if sort:
            if ' ' in sort:
                params['sort'] = f'({sort})'
            else:
                params['sort'] = f'({sort} asc)'

        return self.cupi_api(target_endpoint, data=params)

    # wrapper for api PUT
    def put(self, target_endpoint, put_data):
        return self.cupi_api(target_endpoint, method='PUT', data=put_data)

    # wrapper for api POST
    def post(self, target_endpoint, post_data):
        return self.cupi_api(target_endpoint, method='POST', data=post_data)

    # wrapper for api DELETE
    def delete(self, target_endpoint):
        return self.cupi_api(target_endpoint, method='DELETE')

    def get_prompt(self, greeting_stream_file_uri, filename="greeting.wav"):
        try:
            folder = greeting_stream_file_uri.split("/")[4]
            try:
                print(f"Creating folder {folder}...")
                os.mkdir(folder, 0o0775)
            except Exception as err:
                if not str(err).find('already exists') > -1:
                    print(f"Could not create directory - {err}", file=sys.stderr)

            prompt_file = open(f"{folder}{os.sep}{filename}", "wb")
            wav_content = requests.get(f"https://{self.ipaddr}{greeting_stream_file_uri}/audio",
                                       auth=self.auth, verify=False)
            prompt_file.write(wav_content.content)
            prompt_file.close()

        except Exception as err:
            print(f"Could not download prompt - {err}", file=sys.stderr)

        else:
            return filename

    def get_all_prompts(self, search):
        print(f"Searching for User extension {search}")
        users = self.cupi_api(f"users?query=(%20DtmfAccessId%20is%20{search}%20)")
        if users and len(users) == 0:
            print(f"Searching for User Alias {search}")
            users = self.cupi_api(f"users?query=(%20Alias%20is%20{search}%20)")
        if users and len(users) == 0:
            print(f"Searching for User Display Name {search}")
            users = self.cupi_api(f"users?query=(%20DisplayName%20is%20{search}%20)")

        if users and len(users) > 0:
            ch = self.cupi_api(users[0]['CallhandlerURI'])
            greetings = self.cupi_api(ch[0]['GreetingsURI'])
            for greeting in greetings:
                prompt = self.cupi_api(greeting['GreetingStreamFilesURI'])
                if len(prompt) > 0:
                    prompt_name = prompt[0]['GreetingStreamFileLanguageURI'].split("/")[6]
                    print(f"\tDownloading Prompt {prompt_name}")
                    self.get_prompt(prompt[0]['GreetingStreamFileLanguageURI'], f"{prompt_name}.wav")
            return True

        else:
            print(f"Couldnt find User {search} by extension, searching by name..", file=sys.stderr)
            handlers = self.cupi_api(f"handlers")
            greetings = False
            for handler in handlers:
                if handler['DisplayName'] == search or ('DtmfAccessId' in handler.keys() and handler["DtmfAccessId"] == search):
                    call_handler = self.cupi_api(handler['URI'])
                    if len(call_handler) > 0:
                        greetings = self.cupi_api(call_handler[0]['GreetingsURI'])
                    break
            if not greetings:
                print(f"Couldnt find {search} by name, searching Call Handlers by extension..", file=sys.stderr)
                handlers = self.cupi_api(f"handlers/callhandlers")
                greetings = False
                for handler in handlers:
                    if handler['DisplayName'] == search or ('DtmfAccessId' in handler.keys() and handler["DtmfAccessId"] == search):
                        call_handler = self.cupi_api(handler['URI'])
                        if len(call_handler) > 0:
                            greetings = self.cupi_api(call_handler[0]['GreetingsURI'])
                        break

            if greetings and len(greetings) > 0:
                print(f"Found Call Handler for {search}")
                for greeting in greetings:
                    prompt = self.cupi_api(greeting['GreetingStreamFilesURI'])
                    if len(prompt) > 0:
                        prompt_name = prompt[0]['GreetingStreamFileLanguageURI'].split("/")[6]
                        print(f"\tDownloading {prompt_name}")
                        self.get_prompt(prompt[0]['GreetingStreamFileLanguageURI'], f"{prompt_name}.wav")
                return True
            else:
                print("... and I stiiiiill havent fouuuund what Im looking foooor..")
                return False



import requests
from requests.auth import HTTPBasicAuth
import sys
import json
import urllib
import socket

requests.packages.urllib3.disable_warnings()
try:
    requests.packages.urllib3.contrib.pyopenssl.util.ssl_.DEFAULT_CIPHERS += ':HIGH:!DH:!aNULL'
    requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += ':HIGH:!DH:!aNULL'
except AttributeError:
    # no pyopenssl support used / needed / available
    pass


class Connect:
    def __init__(self, ipaddr, username, passwd):
        self.ipaddr = ipaddr
        self.username = username
        self.auth = HTTPBasicAuth(username, passwd)
        if self.open_port(ipaddr, 443) is False:
            raise Exception (f"EXP Connection Error: {ipaddr}:443 not reachable or open. Is this Expressway?")
        try:
            self.version = self.get("/sysinfo")['SoftwareVersion'].replace('X', '')
        except Exception as err:
            raise Exception (f"EXP Connection Error: API request not valid. Improper credentials?: {err}")

    @staticmethod
    def open_port(ip, port, return_object=[]):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(3)
        try:
            s.connect((ip, int(port)))
            s.close()
            return_object.append(str(port))
            return port
        except Exception as e:
            s.close()
            return False

    # standard REST interface
    def rest(self, method, target_uri=None, params={}):
        if target_uri.find("api/provisioning") > -1:
            target_uri = "".join(target_uri.split("api/provisioning")[1:])

        target_uri = f"https://{self.ipaddr}/api/provisioning{target_uri}"

        if len(params) > 0 and (method == "get" or method == "delete"):
            target_uri += f"?{urllib.parse.urlencode(params)}"

        try:
            # send API request
            response = requests.request(method,
                                        target_uri,
                                        headers={'Accept': 'application/json',
                                                 'Content-Type': 'application/json'},
                                        auth=self.auth,
                                        verify=False,
                                        data=json.dumps(params))

        except Exception as err:
            print(f"Error requesting Expressway API: GET:{target_uri} - {err}", file=sys.stderr)
            return False

        else:
            if 200 <= response.status_code <= 300:
                # Attempt to parse json to dict
                try:
                    result = json.loads(response.text)
                except Exception as err:
                    print(f"EXP response parsing error: {err} - {target_uri} - {response.text}", file=sys.stderr)
                    return False
                else:
                    return result
            elif 400 <= response.status_code <= 600:
                print(f"EXP response parsing error: {target_uri} - {response}", file=sys.stderr)
                return False

    def get(self, target_uri, params={}):
        return self.rest("get", target_uri, params)

    def post(self, target_uri, params={}):
        return self.rest("get", target_uri, params)

    def put(self, target_uri, params={}):
        return self.rest("get", target_uri, params)

    def delete(self, target_uri, params={}):
        return self.rest("get", target_uri, params)

    def patch(self, target_uri, params={}):
        return self.rest("get", target_uri, params)
